﻿using System;
using System.IO;
using System.Text;
using System.Collections.Generic;
using ResTool.Properties;

namespace ResTool {

    class Program {

        //resource file metrics
        struct FileMetrics {

            //starting block
            public int Block;

            //filesize
            public int Size;

            public FileMetrics(int block, int size) {
                Block = block;
                Size = size;
            }
        }

        //headers sigs
        static string entrSig = "ENTR";
        static string bitmapSig = "BM";
        static string resourceSig = "RSCE";

        //define some constants
        static int blockSize = 512;
        static int maxFilenameLength = 256;

        //resource file params
        static byte[] buffer;
        static FileStream resourceImgStream;
        static List<KeyValuePair<String, FileMetrics>> fileList = new List<KeyValuePair<string, FileMetrics>>();

        //read number of bytes relative to offset
        static void readBytes(long offset, long byteCount) {
            if (offset != -1) resourceImgStream.Position = offset;
            buffer = new byte[byteCount];
            resourceImgStream.Read(buffer, 0, (int)byteCount);
        }

        //exit if condition is not met
        static void exitIfNotEqual(bool condition, string errorMessage) {
            if (!condition) {
                Console.WriteLine(errorMessage);
                Environment.Exit(0);
            }
        }

        //app main entry point
        static void Main(string[] args) {

            //cmd-line params
            string switchPack = "-pack";
            string switchUnpack = "-unpack";

            //check cmd-line params
            exitIfNotEqual(args.Length != 0, Resources.RES_TOOL_INFO + '\n' + Resources.RES_TOOL_USAGE);

            //parse cmd-line params
            if (args[0].Equals(switchUnpack))
                UnPackFiles(args.Length > 1 ? args[1] : String.Empty);
            else if (args[0].Equals(switchPack))
                PackFiles(args);
            else {
                Console.WriteLine(Resources.RES_TOOL_INFO);
                Console.WriteLine(Resources.RES_TOOL_USAGE);
            }
        }

        //unpack files from resource image file
        static void UnPackFiles(string resImage) {

            // 0xD0 0x0D 0xFE 0xED (little Endian!)
            UInt32 dtbMagicPacket = 0xedfe0dd0;

            Console.WriteLine(Resources.RES_TOOL_INFO);
            exitIfNotEqual(!resImage.Equals(String.Empty), Resources.RES_FILE_NOT_PROVIDED);

            //check resource image file before opening it
            exitIfNotEqual(File.Exists(resImage), String.Format(Resources.RES_FILE_DONT_EXIST, resImage));

            //open resource image file
            Console.Write(String.Format(Resources.ANALYZING_RES_IMG_FILE, resImage));
            resourceImgStream = File.OpenRead(resImage);

            //read resource file header
            readBytes(0, sizeof(int));

            //validate resource file header
            exitIfNotEqual(Encoding.UTF8.GetString(buffer).Equals(resourceSig), Resources.RES_FILE_HEADER_MISMATCH);

            //read files list
            int entrBlockOffset = blockSize;
            readBytes(entrBlockOffset, entrSig.Length);

            //validate filename headers
            while (Encoding.UTF8.GetString(buffer).Equals(entrSig)) {

                //read filename
                readBytes(entrBlockOffset + entrSig.Length, maxFilenameLength);
                string fileName = Encoding.UTF8.GetString(buffer).TrimEnd('\0');

                //read file starting block
                readBytes(-1, sizeof(int));
                int startBlock = BitConverter.ToInt32(buffer, 0);

                //read filesize
                readBytes(-1, sizeof(int));
                int fileSize = BitConverter.ToInt32(buffer, 0);
                fileList.Add(new KeyValuePair<string, FileMetrics>(fileName, new FileMetrics(startBlock, fileSize)));

                //read another file block
                entrBlockOffset += blockSize;
                readBytes(entrBlockOffset, sizeof(int));
            }

            //validate file list count
            exitIfNotEqual(fileList.Count > 0, Resources.NO_FILES_INSIDE_IMG);

            Console.WriteLine(String.Format(Resources.COUNT_FILES_FOUND, fileList.Count, fileList.Count == 1 ? String.Empty : "s"));

            foreach (var file in fileList) {
                //if file is in a subfolder, create it first 
                if (file.Key.Contains("/"))
                    Directory.CreateDirectory(Path.GetDirectoryName(file.Key));

                using (FileStream fileStream = new FileStream(file.Key, FileMode.Create)) {

                    //validate file headers
                    readBytes(file.Value.Block * blockSize, sizeof(int));
                    if (BitConverter.ToUInt32(buffer, 0) != dtbMagicPacket &&
                        !Encoding.UTF8.GetString(buffer).StartsWith(bitmapSig))
                            Console.WriteLine(String.Format(Resources.FILE_HEADER_NOT_RECOGNIZED, file.Key));

                    //save resource files to file system
                    Console.Write(String.Format(Resources.UNPACKING_FILE_BYTES, file.Key, file.Value.Size));
                    readBytes(file.Value.Block * blockSize, file.Value.Size);
                    fileStream.Write(buffer, 0, file.Value.Size);
                    Console.WriteLine("OK");
                }
            }

            //exit with message
            Console.WriteLine(Resources.REACHED_END_OF_FILE);
        }

        //pack files into resource image file
        static void PackFiles(string[] filenames) {

            //define default resource image filename
            string defaultResImgName = "resource.img";

            Console.WriteLine(Resources.RES_TOOL_INFO);
            exitIfNotEqual(filenames.Length > 1, Resources.NO_FILES_TO_PACK);

            //check if all input filenames are valid
            for (int count = 1; count < filenames.Length; count++)
                exitIfNotEqual(File.Exists(filenames[count]), Resources.FILENAMES_LIST_NOT_VALID);

            //overwrite or create a new resource image file
            Console.Write(String.Format(Resources.CREATING_NEW_RES_FILE, defaultResImgName));
            resourceImgStream = File.Create(defaultResImgName);
            Console.WriteLine("OK");

            //write RSCE header
            Console.Write(Resources.WRITING_RES_IMG_HEADER);
            resourceImgStream.Write(Encoding.UTF8.GetBytes(resourceSig), 0, sizeof(int));

            //write "standard" PTN version
            resourceImgStream.Write(new byte[sizeof(short)], 0, sizeof(short));

            //write "standard" TBL version
            resourceImgStream.Write(new byte[sizeof(short)], 0, sizeof(short));

            //write block sizes and offsets (?)
            resourceImgStream.Write(new byte[] { 1, 1, 1, 0 }, 0, sizeof(int));

            //write number of blocks/files
            resourceImgStream.Write(BitConverter.GetBytes(filenames.Length-1), 0, sizeof(int));

            Console.WriteLine("OK");

            //write zero-bytes gap until first filename entry offset
            int zeroBytesGap = blockSize - (int)resourceImgStream.Position;
            resourceImgStream.Write(new byte[zeroBytesGap], 0, zeroBytesGap);

            //offset of first filename
            int entrBlockOffset = blockSize;

            //initial resource image filesize (only header, without files themselves)
            int resImgFileSize = blockSize * filenames.Length;

            //write filename headers
            for (int count = 1; count < filenames.Length; count++) {

                //write filename entry header
                FileInfo fileInfo = new FileInfo(filenames[count]);
                resourceImgStream.Write(Encoding.UTF8.GetBytes(entrSig), 0, sizeof(int));

                //write filename with extension
                buffer = new byte[maxFilenameLength];
                string filePath = fileInfo.ToString().Replace('\\','/');
                Array.Copy(Encoding.UTF8.GetBytes(filePath), buffer, filePath.Length);
                resourceImgStream.Write(buffer, 0, maxFilenameLength);

                //write file block position
                resourceImgStream.Write(BitConverter.GetBytes(resImgFileSize / blockSize), 0, sizeof(int));

                //write filesize
                int fileSize = (int)fileInfo.Length;
                resourceImgStream.Write(BitConverter.GetBytes(fileSize), 0, sizeof(int));

                //round resource image filesize to nearest bigger block size
                resImgFileSize += (int)Math.Ceiling(fileSize / (blockSize*1f)) * blockSize;
                
                //set offset to another file block
                entrBlockOffset += blockSize;

                //write zero-bytes gap until the next file block
                zeroBytesGap = entrBlockOffset - (int)resourceImgStream.Position;
                resourceImgStream.Write(new byte[zeroBytesGap], 0, zeroBytesGap);
            }

            Console.WriteLine(String.Format(Resources.PACKING_FILE_PROGRESS, filenames.Length-1, filenames.Length-1 == 1 ? String.Empty : "s"));

            //pack all input files
            for (int count = 1; count < filenames.Length; count++) {

                string filename = filenames[count];
                Console.Write(filename + "... ");

                //copy files from file system into resource file
                byte[] fileBytes = File.ReadAllBytes(filename);
                buffer = new byte[(int)Math.Ceiling(fileBytes.Length / (blockSize*1f)) * blockSize];
                Array.Copy(fileBytes, buffer, fileBytes.Length);
                resourceImgStream.Write(buffer, 0, buffer.Length);

                Console.WriteLine("OK");
            }

            //close new resource file and exit with message
            resourceImgStream.Close();
            Console.WriteLine(Resources.PACKING_FILES_COMPLETED);
        }
    }
}
